# This file should be saved into one of the config directories provided by `jupyter --path`.

c.JupyterLabPioneerApp.exporters = [
    {
        # sends telemetry data to a remote http endpoint (AWS S3 bucket)
        "type": "remote_exporter",
        "args": {
            "id": "S3Exporter",
            "url": "https://telemetry.mentoracademy.org/telemetry-edtech-labs-si-umich-edu/dev/test-telemetry",
            "env": ["WORKSPACE_ID"],
        },
    },
    # {
    #     # sends telemetry data to a remote http endpoint (an AWS Lambda function that exports data to MongoDB)
    #     "type": "remote_exporter",
    #     "args": {
    #         "id": "MongoDBLambdaExporter",
    #         "url": "https://68ltdi5iij.execute-api.us-east-1.amazonaws.com/mongo",
    #         "params": {
    #             "mongo_cluster": "mengyanclustertest.6b83fsy.mongodb.net",
    #             "mongo_db": "telemetry",
    #             "mongo_collection": "dev",
    #         },
    #         "env": ["WORKSPACE_ID"],
    #     },
    # },
    {
        # sends telemetry data to a remote http endpoint (an AWS Lambda function that exports data to InfluxDB)
        "type": "remote_exporter",
        "args": {
            "id": "InfluxDBLambdaExporter",
            "url": "https://68ltdi5iij.execute-api.us-east-1.amazonaws.com/influx",
            "params": {
                "influx_bucket": "telemetry_dev",
                "influx_measurement": "si101_fa24",
            },
        },
        "activeEvents": [
            {"name": "CellEditEvent", "logWholeNotebook": False},
        ],  # exporter's local active_events config will override global activeEvents config
    },
]

c.JupyterLabPioneerApp.activeEvents = [
    {"name": "ActiveCellChangeEvent", "logWholeNotebook": False},
    {"name": "CellAddEvent", "logWholeNotebook": False},
    # {"name": "CellEditEvent", "logWholeNotebook": False},
    {"name": "CellExecuteEvent", "logWholeNotebook": False},
    {"name": "CellRemoveEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCopyEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCutEvent", "logWholeNotebook": False},
    # {"name": "ClipboardPasteEvent", "logWholeNotebook": False},
    # {"name": "NotebookHiddenEvent", "logWholeNotebook": False},
    # {"name": "NotebookOpenEvent", "logWholeNotebook": False},
    # {"name": "NotebookSaveEvent", "logWholeNotebook": False},
    # {"name": "NotebookScrollEvent", "logWholeNotebook": False},
    # {"name": "NotebookVisibleEvent", "logWholeNotebook": False},
]



import sys
import os

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath('../..'))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'jupyterlab-pioneer'
copyright = '2023, University of Michigan'
author = 'University of Michigan'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo',
    'sphinx.ext.napoleon',
    'sphinx_rtd_theme',
    'sphinxcontrib.mermaid'
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

language = 'en'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'analytics_anonymize_ip': False,
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    'style_nav_header_background': 'white',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False
}
html_static_path = ['_static']

# -- Options for todo extension ----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/todo.html#configuration

todo_include_todos = True



"""This module provides 5 default exporters for the extension. If the exporter function name is mentioned in the configuration file or in the notebook metadata, the extension will use the corresponding exporter function when the jupyter lab event is fired.

Attributes:
    default_exporters: a map from function names to callable exporter functions::

        default_exporters: dict[str, Callable[[dict], dict or Awaitable[dict]]] = {
            "console_exporter": console_exporter,
            "command_line_exporter": command_line_exporter,
            "file_exporter": file_exporter,
            "remote_exporter": remote_exporter,
            "opentelemetry_exporter": opentelemetry_exporter,
        }
"""

import json
import os
import datetime
from collections.abc import Callable, Awaitable
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.httputil import HTTPHeaders
from tornado.escape import to_unicode


def console_exporter(args: dict) -> dict:
    """This exporter sends telemetry data to the browser console.

    Args:
        args(dict): arguments to pass to the exporter function, defined in the configuration file (except 'data', which is gathered by the extension). It has the following structure:
            ::

                {
                    'id': # (optional) exporter id,
                    'data': # telemetry data
                }

    Returns:
        dict:
            ::

                {
                    'exporter': # exporter id or 'ConsoleExporter',
                    'message': # telemetry data
                }

    """

    return {"exporter": args.get("id") or "ConsoleExporter", "message": args["data"]}


def command_line_exporter(args: dict) -> dict:
    """This exporter sends telemetry data to the python console jupyter is running on.

    Args:
        args (dict): arguments to pass to the exporter function, defined in the configuration file (except 'data', which is gathered by the extension). It has the following structure:
            ::

                {
                    'id': # (optional) exporter id,
                    'data': # telemetry data
                }

    Returns:
        dict:
            ::

                {
                    'exporter': # exporter id or 'CommandLineExporter',
                }
    
    """

    print(args["data"])
    return {
        "exporter": args.get("id") or "CommandLineExporter",
    }


def file_exporter(args: dict) -> dict:
    """This exporter writes telemetry data to local file.

    Args:
        args (dict): arguments to pass to the exporter function, defined in the configuration file (except 'data', which is gathered by the extension). It has the following structure:
            ::

                {
                    'id': # (optional) exporter id,
                    'path': # local file path,
                    'data': # telemetry data
                }

    Returns:
        dict:
            ::

                {
                    'exporter': # exporter id or 'FileExporter',
                }
    """

    with open(args.get("path"), "a+", encoding="utf-8") as f:
        json.dump(args["data"], f, ensure_ascii=False, indent=4)
        f.write(",")
    return {
        "exporter": args.get("id") or "FileExporter",
    }


async def remote_exporter(args: dict) -> dict:
    """This exporter sends telemetry data to a remote http endpoint.

    Args:
        args (dict): arguments to pass to the exporter function, defined in the configuration file (except 'data', which is gathered by the extension). It has the following structure:
            ::

                {
                    'id': # (optional) exporter id,
                    'url': # http endpoint url,
                    'params': # (optional) additional parameters to pass to the http endpoint,
                    'env': # (optional) environment variables to pass to the http endpoint,
                    'data': # telemetry data
                }

    Returns:
        dict:
            ::

                {
                    'exporter': exporter id or 'RemoteExporter',
                    'message': {
                        'code': http response code,
                        'reason': http response reason,
                        'body': http response body
                    }
                }

    """
    http_client = AsyncHTTPClient()
    unix_timestamp = args["data"].get("eventDetail").get("eventTime")
    utc_datetime = datetime.datetime.fromtimestamp(unix_timestamp/1000.0, tz=datetime.timezone.utc)
    url = args.get("url")
    if "s3" in args.get("id").lower():
        url = "%s/%d/%d/%d/%d" % (args.get("url"), utc_datetime.year, utc_datetime.month, utc_datetime.day, utc_datetime.hour)
    request = HTTPRequest(
        url=url,
        method="POST",
        body=json.dumps(
            {
                "data": args["data"],
                "params": args.get(
                    "params"
                ),  # none if exporter does not contain 'params'
                "env": [{x: os.getenv(x)} for x in args.get("env")]
                if (args.get("env"))
                else [],
            }
        ),
        headers=HTTPHeaders({"content-type": "application/json"}),
    )
    response = await http_client.fetch(request, raise_error=False)
    return {
        "exporter": args.get("id") or "RemoteExporter",
        "message": {
            "code": response.code,
            "reason": response.reason,
            "body": to_unicode(response.body),
        },
    }

def opentelemetry_exporter(args: dict) -> dict:
    """This exporter sends telemetry data via otlp

    """
    from opentelemetry import trace

    current_span = trace.get_current_span()
    event_detail = args['data']['eventDetail']
    notebook_state = args['data']['notebookState']
    attributes = {
        "notebookSessionId": notebook_state['sessionID'],
        'notebookPath': notebook_state['notebookPath'],
        "event": event_detail['eventName']
    }
    current_span.add_event(event_detail['eventName'], attributes=attributes)

    return {
        "exporter": args.get("id") or "OpenTelemetryExporter",
    }

default_exporters: "dict[str, Callable[[dict], dict or Awaitable[dict]]]" = {
    "console_exporter": console_exporter,
    "command_line_exporter": command_line_exporter,
    "file_exporter": file_exporter,
    "remote_exporter": remote_exporter,
    "opentelemetry_exporter": opentelemetry_exporter,
}



# This file should be saved into one of the config directories provided by `jupyter --path`.

c.JupyterLabPioneerApp.exporters = [
    {
        # sends telemetry data to the browser console
        "type": "console_exporter",
    },
    {
        # sends telemetry data to the python console jupyter is running on
        "type": "command_line_exporter",
    },
]

c.JupyterLabPioneerApp.activeEvents = [
    {"name": "ActiveCellChangeEvent", "logWholeNotebook": False},
    {"name": "CellAddEvent", "logWholeNotebook": False},
    # {"name": "CellEditEvent", "logWholeNotebook": False},
    {"name": "CellExecuteEvent", "logWholeNotebook": False},
    {"name": "CellRemoveEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCopyEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCutEvent", "logWholeNotebook": False},
    # {"name": "ClipboardPasteEvent", "logWholeNotebook": False},
    # {"name": "NotebookHiddenEvent", "logWholeNotebook": False},
    # {"name": "NotebookOpenEvent", "logWholeNotebook": False},
    # {"name": "NotebookSaveEvent", "logWholeNotebook": False},
    # {"name": "NotebookScrollEvent", "logWholeNotebook": False},
    # {"name": "NotebookVisibleEvent", "logWholeNotebook": False},
]


# This file should be saved into one of the config directories provided by `jupyter --path`.

def my_custom_exporter(args):
    # write your own exporter logic here
    return {
        "exporter": args.get("id"),
        "message": ""
    }

c.JupyterLabPioneerApp.exporters = [
    {
        "type": "custom_exporter",
        "args": {
            "id": "MyCustomExporter"
            # add additional args for your exporter function here
        },
    }
]

c.JupyterLabPioneerApp.custom_exporter = {
    'MyCustomExporter': my_custom_exporter,
}

c.JupyterLabPioneerApp.activeEvents = [
    {"name": "ActiveCellChangeEvent", "logWholeNotebook": False},
    {"name": "CellAddEvent", "logWholeNotebook": False},
    # {"name": "CellEditEvent", "logWholeNotebook": False},
    {"name": "CellExecuteEvent", "logWholeNotebook": False},
    {"name": "CellRemoveEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCopyEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCutEvent", "logWholeNotebook": False},
    # {"name": "ClipboardPasteEvent", "logWholeNotebook": False},
    # {"name": "NotebookHiddenEvent", "logWholeNotebook": False},
    # {"name": "NotebookOpenEvent", "logWholeNotebook": False},
    # {"name": "NotebookSaveEvent", "logWholeNotebook": False},
    # {"name": "NotebookScrollEvent", "logWholeNotebook": False},
    # {"name": "NotebookVisibleEvent", "logWholeNotebook": False},
]



# This file should be saved into one of the config directories provided by `jupyter --path`.

"""An array of the exporters.

This extension provides 5 default exporters:
`console_exporter`, which sends telemetry data to the browser console.
`command_line_exporter`, which sends telemetry data to the python console jupyter is running on.
`file_exporter`, which saves telemetry data to local file.
`remote_exporter`, which sends telemetry data to a remote http endpoint.
`opentelemetry_exporter`, which sends telemetry data via otlp.

Additionally, users can import default exporters or write customized exporters in the configuration file.
"""
c.JupyterLabPioneerApp.exporters = [
    {
        # sends telemetry data to the browser console
        "type": "console_exporter",
    },
    {
        # sends telemetry data to the python console jupyter is running on
        "type": "command_line_exporter",
    },
    {
        # writes telemetry data to local file
        "type": "file_exporter",
        "args": {
            "path": "log"
        },
    },
    {
        # sends telemetry data to a remote http endpoint (AWS S3 bucket)
        "type": "remote_exporter",
        "args": {
            "id": "S3Exporter",
            "url": "https://telemetry.mentoracademy.org/telemetry-edtech-labs-si-umich-edu/dev/test-telemetry",
            "env": ["WORKSPACE_ID"],
        },
    },
    {
        # sends telemetry data to a remote http endpoint (an AWS Lambda function that exports data to MongoDB)
        "type": "remote_exporter",
        "args": {
            "id": "MongoDBLambdaExporter",
            "url": "https://68ltdi5iij.execute-api.us-east-1.amazonaws.com/mongo",
            "params": {
                "mongo_cluster": "mengyanclustertest.6b83fsy.mongodb.net",
                "mongo_db": "telemetry",
                "mongo_collection": "dev",
            },
            "env": ["WORKSPACE_ID"],
        },
    },
    {
        # sends telemetry data to a remote http endpoint (an AWS Lambda function that exports data to InfluxDB)
        "type": "remote_exporter",
        "args": {
            "id": "InfluxDBLambdaExporter",
            "url": "https://68ltdi5iij.execute-api.us-east-1.amazonaws.com/influx",
            "params": {
                "influx_bucket": "telemetry_dev",
                "influx_measurement": "si101_fa24",
            },
        },
        "activeEvents": [
            {"name": "CellEditEvent", "logWholeNotebook": False},
        ],  # exporter's local active_events config will override global activeEvents config
    },
]

"""An array of active events.
This is a global config. It could be override by `activeEvents` defined inside of individual exporter configs
The extension would only generate and export data for valid events (
    1. that have an id associated with the event class, 
    2. the event name is included in `activeEvents`
).
The extension will export the entire notebook content only for valid events with the logWholeNotebook flag == True
"""
c.JupyterLabPioneerApp.activeEvents = [
    {"name": "ActiveCellChangeEvent", "logWholeNotebook": False},
    {"name": "CellAddEvent", "logWholeNotebook": False},
    # {"name": "CellEditEvent", "logWholeNotebook": False},
    {"name": "CellExecuteEvent", "logWholeNotebook": False},
    {"name": "CellRemoveEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCopyEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCutEvent", "logWholeNotebook": False},
    # {"name": "ClipboardPasteEvent", "logWholeNotebook": False},
    # {"name": "NotebookHiddenEvent", "logWholeNotebook": False},
    # {"name": "NotebookOpenEvent", "logWholeNotebook": False},
    # {"name": "NotebookSaveEvent", "logWholeNotebook": False},
    # {"name": "NotebookScrollEvent", "logWholeNotebook": False},
    # {"name": "NotebookVisibleEvent", "logWholeNotebook": False},
]



try:
    from ._version import __version__
except ImportError:
    import warnings
    warnings.warn("Importing 'jupyterlab_pioneer' outside a proper installation.")
    __version__ = "dev"

from .application import JupyterLabPioneerApp

def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": "jupyterlab-pioneer"
    }]

def _jupyter_server_extension_points():
    return [{
        "module": "jupyterlab_pioneer",
        "app": JupyterLabPioneerApp
    }]

load_jupyter_server_extension = JupyterLabPioneerApp.load_classic_server_extension



"""This module defines the extra request handlers the pioneer extension needs
"""
import os
import json
import inspect
import tornado
from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin
from ._version import __version__
from .default_exporters import default_exporters

class RouteHandler(ExtensionHandlerMixin, JupyterHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # The following decorator should be present on all verb methods (head, get, post,
    # patch, put, delete, options) to ensure only authorized user can request the
    # Jupyter server
    @tornado.web.authenticated
    def get(self, resource):
        """GET method

        Args:
            resource (str): the name of the resource requested. It is expected to be one of "version", "environ", or "config".
        
        Returns:
            str(json):
                If resource is "version", the server responses with a json serialized obj of the version string.

                If resource is "environ", the server responses with a json serialized obj of current environment variables.

                If resource is "config", the server responses with a json serialized obj containing "activeEvents" and "exporters" configurations from the configuration file.

                For other resources, set the status code to 404 not found.
        
        """
        try:
            self.set_header("Content-Type", "application/json")
            if resource == "version":
                self.finish(json.dumps(__version__))
            elif resource == "environ":
                self.finish(json.dumps(dict(os.environ.items())))
            elif resource == "config":
                self.finish(
                    json.dumps(
                        {
                            "activeEvents": self.extensionapp.activeEvents,
                            "exporters": self.extensionapp.exporters,
                        }
                    )
                )
            else:
                self.set_status(404)
        except Exception as e:
            self.log.error(str(e))
            self.set_status(500)
            self.finish(json.dumps(str(e)))

    @tornado.web.authenticated
    async def post(self, resource):
        """POST method

        Args:
            resource (str): the name of the resource requested. It is expected to be "export".
        
        Returns:
            str(json):
            If resource is "export", the server calls the asynchronous export function :func:`export`, and responses with the json serialized export result.
            
            For other resources, set the status code to 404 not found.
        """
        try:
            if resource == "export":
                result = await self.export()
                self.finish(json.dumps(result))
            else:
                self.set_status(404)

        except Exception as e:
            self.log.error(str(e))
            self.set_status(500)
            self.finish(json.dumps(str(e)))

    async def export(self):
        """This function exports telemetry data with requested exporters.

        The function first parse the request body to get the event data and the corresponding exporter requested for the event. 
        Then base on the exporter type, the function either calls the default exporters or tries to access the custom exporter defined in the configuration file. 

        Returns:
            dict:
                ::

                    {
                        "exporter": # exporter type,
                        "message": # execution message of the exporter function
                    }
        """
        body = json.loads(self.request.body)
        exporter = body.get("exporter")
        data = {
            "eventDetail": body.get("eventDetail"),
            "notebookState": body.get("notebookState"),
        }
        exporter_type = exporter.get("type")
        args = exporter.get("args") or {}  # id, url, path, params, env
        args["data"] = data
        if exporter_type in default_exporters:
            exporter_func = default_exporters[exporter_type]
            if inspect.iscoroutinefunction(exporter_func):
                result = await exporter_func(args)
            else:
                result = exporter_func(args)
            return result
        if exporter_type == "custom_exporter":
            custom_exporter = self.extensionapp.custom_exporter
            if (
                custom_exporter
                and args.get("id") in custom_exporter
                and custom_exporter.get(args.get("id"))
            ):
                exporter_func = custom_exporter.get(args.get("id"))
                if inspect.iscoroutinefunction(exporter_func):
                    result = await exporter_func(args)
                else:
                    result = exporter_func(args)
                return result
            else:
                return {
                    "exporter": exporter_type,
                    "message": "[Error] custom exporter is not defined",
                }
        return {
            "exporter": exporter_type,
            "message": "[Error] exporter is not supported",
        }



"""This module defines the extension app name, reads configurable variables from the configuration file, and adds the extra request handlers from :mod:`.handlers` module to Jupyter Server's Tornado Web Application.
"""

from traitlets import List, Dict
from jupyter_server.extension.application import ExtensionApp
from .handlers import RouteHandler

class JupyterLabPioneerApp(ExtensionApp):
    name = "jupyterlab_pioneer"
    """Extension app name"""

    activeEvents = List([]).tag(config=True)
    """An array of active events defined in the configuration file.

    Global config, could be override by `activeEvents` defined inside of individual exporter configs.

    The extension would only generate and export data for valid events (
    1. that have an id associated with the event class, 
    2. the event name is included in `activeEvents`
    ).
    
    The extension will export the entire notebook content only for valid events when the `logWholeNotebook` flag is `True`.

    Examples:
        ::
        
            # in the configuration file
            c.JupyterLabPioneerApp.activeEvents = 
            [
                {"name": "ActiveCellChangeEvent", "logWholeNotebook": False},
                {"name": "CellAddEvent", "logWholeNotebook": False},
                # {"name": "CellEditEvent", "logWholeNotebook": False},
                {"name": "CellExecuteEvent", "logWholeNotebook": False},
                {"name": "CellRemoveEvent", "logWholeNotebook": False},
                # {"name": "ClipboardCopyEvent", "logWholeNotebook": False},
                # {"name": "ClipboardCutEvent", "logWholeNotebook": False},
                # {"name": "ClipboardPasteEvent", "logWholeNotebook": False},
                # {"name": "NotebookHiddenEvent", "logWholeNotebook": False},
                # {"name": "NotebookOpenEvent", "logWholeNotebook": False},
                # {"name": "NotebookSaveEvent", "logWholeNotebook": False},
                # {"name": "NotebookScrollEvent", "logWholeNotebook": False},
                # {"name": "NotebookVisibleEvent", "logWholeNotebook": False},
            ]
    """

    exporters = List([]).tag(config=True)
    """ An array of the exporters defined in the configuration file.

    This extension provides 5 default exporters in the :mod:`.default_exporters` module:

    :func:`.default_exporters.console_exporter`, which sends telemetry data to the browser console.

    :func:`.default_exporters.command_line_exporter`, which sends telemetry data to the python console jupyter is running on.

    :func:`.default_exporters.file_exporter`, which saves telemetry data to local file.

    :func:`.default_exporters.remote_exporter`, which sends telemetry data to a remote http endpoint.

    :func:`.default_exporters.opentelemetry_exporter`, which sends telemetry data via otlp.

    Additionally, users can import default exporters or write customized exporters in the configuration file.

    Examples:
        ::

            # in the configuration file
            c.JupyterLabPioneerApp.exporters = 
            [
                {
                    "type": "console_exporter",
                },
                {
                    "type": "command_line_exporter",
                },
            ]
    """

    custom_exporter = Dict({}).tag(config=True)
    """A dictionary of custom exporter defined in the configuration file
    
    Examples:
        ::
        
            # in the configuration file
                def my_custom_exporter(args):
                    # write your own exporter logic here
                    return {
                        "exporter": args.get("id"),
                        "message": ""
                    }

                c.JupyterLabPioneerApp.exporters = [
                    {
                        "type": "custom_exporter",
                        "args": {
                            "id": "MyCustomExporter"
                            # add additional args for your exporter function here
                        },
                    }
                ]

                c.JupyterLabPioneerApp.custom_exporter = {
                    'MyCustomExporter': my_custom_exporter,
                }
    """

    def initialize_handlers(self):
        """This function adds the extra request handlers from :mod:`.handlers` module to Jupyter Server's Tornado Web Application.
        """
        try:
            self.handlers.extend([(r"/jupyterlab-pioneer/(.*)", RouteHandler)])
        except Exception as e:
            self.log.error(str(e))
            raise e



# This file should be saved into one of the config directories provided by `jupyter --path`.

c.JupyterLabPioneerApp.exporters = [
    {
        # writes telemetry data to local file
        "type": "file_exporter",
        "args": {
            "path": "log"
        },
    },
]

c.JupyterLabPioneerApp.activeEvents = [
    {"name": "ActiveCellChangeEvent", "logWholeNotebook": False},
    {"name": "CellAddEvent", "logWholeNotebook": False},
    # {"name": "CellEditEvent", "logWholeNotebook": False},
    {"name": "CellExecuteEvent", "logWholeNotebook": False},
    {"name": "CellRemoveEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCopyEvent", "logWholeNotebook": False},
    # {"name": "ClipboardCutEvent", "logWholeNotebook": False},
    # {"name": "ClipboardPasteEvent", "logWholeNotebook": False},
    # {"name": "NotebookHiddenEvent", "logWholeNotebook": False},
    # {"name": "NotebookOpenEvent", "logWholeNotebook": False},
    # {"name": "NotebookSaveEvent", "logWholeNotebook": False},
    # {"name": "NotebookScrollEvent", "logWholeNotebook": False},
    # {"name": "NotebookVisibleEvent", "logWholeNotebook": False},
]
